from numba import cuda, types
import numpy as np
import math

from numpy.ma import count

MAX_DIM = 2
DEFAULT_DTYPE = types.float32
MINIMAL_DISTANCE = 10 ** -18



@cuda.jit
def convert_planets_to_image(global_positions, global_masses, global_window_start, global_window_end, global_image):
    # thread unic values
    tx = cuda.threadIdx.x
    bx = cuda.blockIdx.x
    unique_x = tx + bx * cuda.blockDim.x
   
    width = global_image.shape[1]
    height = global_image.shape[0]
    dim = 2
    index = x_idx, y_idx = unique_x % width, unique_x // width
    count_of_bodies = global_positions.shape[0]

    if unique_x >= width * height:
        return

    local_pixel_position = cuda.local.array(MAX_DIM, DEFAULT_DTYPE)
    local_pixel_color = cuda.local.array(3, DEFAULT_DTYPE) # (r, g, b)

    shared_position = cuda.shared.array(MAX_DIM, DEFAULT_DTYPE)
    shared_mass = cuda.shared.array(1, DEFAULT_DTYPE)

    for i in range(dim):
        local_pixel_color[i] = 0
        step = (global_window_end[i] - global_window_start[i]) / global_image.shape[i]
        local_pixel_position[i] = global_window_start[i] + index[i] * step

    # iterate over all positions
    for n in range(count_of_bodies):
        # load position
        cuda.syncthreads()
        for i in range(dim):
            shared_position[i] = global_positions[n][i]
        shared_mass[0] = global_masses[n]
        cuda.syncthreads()


        # calculate distance
        squared_dist = 0
        for i in range(dim):
            squared_dist += (shared_position[i] + local_pixel_position[i]) ** 2
        squared_dist += MINIMAL_DISTANCE

        # calculate color
        x = squared_dist
        a = shared_mass[0]
        val = a / (x)

        #red star
        local_pixel_color[0] += 0.04 * val
        local_pixel_color[1] += 0.02 * val
        local_pixel_color[2] += 0.01 * val

        # blue star
        # local_pixel_color[0] += 0.01 * val
        # local_pixel_color[1] += 0.02 * val
        # local_pixel_color[2] += 0.04 * val

    for i in range(3):
        local_pixel_color[i] = min(1, max(local_pixel_color[i], 0))
    
    for i in range(3):
        global_image[y_idx][x_idx][i] = local_pixel_color[i]
    


if __name__ == "__main__":
    from PIL import Image
    import numpy as np
    from kernel import blocks_per_grid

    np.set_printoptions(threshold=np.inf)
    
    w_start = np.array([-400, -400], dtype=np.float64)
    w_end = np.array([400, 400], dtype=np.float64)
    positions = np.random.randint(w_start, w_end, (40, 2))
    masses = np.random.random(positions.shape[0]) ** 2 * 20
    im = np.zeros([2000, 2000, 3], dtype=np.float64)
    tpb = 320
    bpg = blocks_per_grid(2000**2, tpb)
    kf = convert_planets_to_image[bpg, tpb]
    import time
    t = time.time()
    kf(positions, masses,  w_start, w_end, im)
    print(time.time() - t)
    image = Image.fromarray((im * 255).astype(np.uint8))
    image.show()